# 1 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr/data/dbcsr_ptr_util.F"
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2018  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   DBCSR pointer and unmanaged array utilities
!> \author  Urban Borstndik
!> \date    2010-02-18
!> \version 0.9
!>
!> <b>Modification history:</b>
!> - 2010-02-18 Moved from dbcsr_util
!> - 2010-06-18 Moved all pointer methods into here.
! **************************************************************************************************
MODULE dbcsr_ptr_util
   USE acc_hostmem,                     ONLY: acc_hostmem_allocate,&
                                              acc_hostmem_deallocate
   USE dbcsr_data_types,                ONLY: dbcsr_data_obj,&
                                              dbcsr_memtype_default,&
                                              dbcsr_memtype_type,&
                                              dbcsr_type_complex_4,&
                                              dbcsr_type_complex_8,&
                                              dbcsr_type_real_4,&
                                              dbcsr_type_real_8
   USE kinds,                           ONLY: dp,&
                                              int_4,&
                                              int_8,&
                                              real_4,&
                                              real_8
   USE message_passing,                 ONLY: mp_allocate,&
                                              mp_deallocate
#include "../../base/base_uses.f90"

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_ptr_util'

   LOGICAL, PARAMETER :: careful_mod = .FALSE.

   PUBLIC :: ensure_array_size
   PUBLIC :: memory_allocate, memory_deallocate
   PUBLIC :: memory_zero
   PUBLIC :: pointer_view
   PUBLIC :: pointer_rank_remap2
   PUBLIC :: memory_copy

   INTERFACE ensure_array_size
      MODULE PROCEDURE ensure_array_size_i, ensure_array_size_l
      MODULE PROCEDURE ensure_array_size_s, ensure_array_size_d, &
         ensure_array_size_c, ensure_array_size_z
   END INTERFACE

   INTERFACE pointer_view
      MODULE PROCEDURE pointer_view_s, pointer_view_d, &
         pointer_view_c, pointer_view_z
      MODULE PROCEDURE pointer_view_i, pointer_view_l
      MODULE PROCEDURE pointer_view_a
   END INTERFACE

   INTERFACE pointer_rank_remap2
      MODULE PROCEDURE pointer_s_rank_remap2, pointer_d_rank_remap2, &
         pointer_c_rank_remap2, pointer_z_rank_remap2, &
         pointer_l_rank_remap2, pointer_i_rank_remap2
   END INTERFACE

   INTERFACE memory_copy
      MODULE PROCEDURE mem_copy_i, mem_copy_l, &
         mem_copy_s, mem_copy_d, &
         mem_copy_c, mem_copy_z
   END INTERFACE

   INTERFACE memory_zero
      MODULE PROCEDURE mem_zero_i, mem_zero_l
      MODULE PROCEDURE mem_zero_s, mem_zero_d, mem_zero_c, mem_zero_z
   END INTERFACE

   INTERFACE memory_allocate
      MODULE PROCEDURE mem_alloc_i, mem_alloc_l, mem_alloc_s, mem_alloc_d, mem_alloc_c, mem_alloc_z
      MODULE PROCEDURE mem_alloc_i_2d, mem_alloc_l_2d, mem_alloc_s_2d, mem_alloc_d_2d, mem_alloc_c_2d, mem_alloc_z_2d
   END INTERFACE

   INTERFACE memory_deallocate
      MODULE PROCEDURE mem_dealloc_i, mem_dealloc_l, mem_dealloc_s, mem_dealloc_d, mem_dealloc_c, mem_dealloc_z
      MODULE PROCEDURE mem_dealloc_i_2d, mem_dealloc_l_2d, mem_dealloc_s_2d, mem_dealloc_d_2d, mem_dealloc_c_2d, mem_dealloc_z_2d
   END INTERFACE

   TYPE dbcsr_data_allocation_type
      LOGICAL :: use_mpi_allocator = .TRUE.
   END TYPE dbcsr_data_allocation_type

   PUBLIC :: dbcsr_data_allocation_type
   TYPE(dbcsr_data_allocation_type), PUBLIC, SAVE :: dbcsr_data_allocation = dbcsr_data_allocation_type() ! defaults


CONTAINS

! **************************************************************************************************
!> \brief Repoints a pointer into a part of a data area
!> \param[in,out] new_area    repoints this encapsulated pointer
!> \param[in] area            area to point into
!> \param[in] offset          point to this offset in area
!> \param[in] len             (optional) length of data area to point to
!> \return copy of new_area
! **************************************************************************************************
   FUNCTION pointer_view_a(new_area, area, offset, len) RESULT(narea2)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: new_area
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: area
      INTEGER, INTENT(IN)                                :: offset
      INTEGER, INTENT(IN), OPTIONAL                      :: len
      TYPE(dbcsr_data_obj)                               :: narea2

      CHARACTER(len=*), PARAMETER :: routineN = 'pointer_view_a', routineP = moduleN//':'//routineN

      IF (area%d%data_type /= new_area%d%data_type) &
         CPABORT("Incompatible data types.")
      IF (PRESENT(len)) THEN
         SELECT CASE (area%d%data_type)
         CASE (dbcsr_type_real_4)
            new_area%d%r_sp => area%d%r_sp(offset:offset+len-1)
         CASE (dbcsr_type_real_8)
            new_area%d%r_dp => area%d%r_dp(offset:offset+len-1)
         CASE (dbcsr_type_complex_4)
            new_area%d%c_sp => area%d%c_sp(offset:offset+len-1)
         CASE (dbcsr_type_complex_8)
            new_area%d%c_dp => area%d%c_dp(offset:offset+len-1)
         CASE default
            CPABORT("Invalid data type.")
         END SELECT
      ELSE
         SELECT CASE (area%d%data_type)
         CASE (dbcsr_type_real_4)
            new_area%d%r_sp => area%d%r_sp(offset:)
         CASE (dbcsr_type_real_8)
            new_area%d%r_dp => area%d%r_dp(offset:)
         CASE (dbcsr_type_complex_4)
            new_area%d%c_sp => area%d%c_sp(offset:)
         CASE (dbcsr_type_complex_8)
            new_area%d%c_dp => area%d%c_dp(offset:)
         CASE default
            CPABORT("Invalid data type.")
         END SELECT
      ENDIF
      narea2 = new_area
   END FUNCTION pointer_view_a

# 1 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr/data/dbcsr_ptr_util.f90" 1
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2018  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

# 1 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr/data/dbcsr.fypp" 1
# 8 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr/data/dbcsr.fypp"
# 39 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr/data/dbcsr.fypp"
# 7 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr/data/dbcsr_ptr_util.f90" 2
# 8 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr/data/dbcsr_ptr_util.f90"
! **************************************************************************************************
!> \brief Returns a pointer with different bounds.
!> \param[in] original   original data pointer
!> \param[in] lb lower and upper bound for the new pointer view
!> \param[in] ub lower and upper bound for the new pointer view
!> \return new pointer
! **************************************************************************************************
  FUNCTION pointer_view_d (original, lb, ub) RESULT (view)
    REAL(kind=real_8), DIMENSION(:), POINTER :: original, view
    INTEGER, INTENT(IN)                  :: lb, ub
    view => original(lb:ub)
  END FUNCTION pointer_view_d


! **************************************************************************************************
!> \brief Ensures that an array is appropriately large.
!> \param[in,out] array       array to verify and possibly resize
!> \param[in] lb    (optional) desired array lower bound
!> \param[in] ub    desired array upper bound
!> \param[in] factor          (optional) factor by which to exagerrate
!>                            enlargements
!> \param[in] nocopy          (optional) copy array on enlargement; default
!>                            is to copy
!> \param[in] memory_type     (optional) use special memory
!> \param[in] zero_pad        (optional) zero new allocations; default is to
!>                            write nothing
! **************************************************************************************************
  SUBROUTINE ensure_array_size_d(array, array_resize, lb, ub, factor,&
       nocopy, memory_type, zero_pad)
    REAL(kind=real_8), DIMENSION(:), POINTER                 :: array
    REAL(kind=real_8), DIMENSION(:), POINTER, OPTIONAL       :: array_resize
    INTEGER, INTENT(IN), OPTIONAL                  :: lb
    INTEGER, INTENT(IN)                            :: ub
    REAL(KIND=dp), INTENT(IN), OPTIONAL            :: factor
    LOGICAL, INTENT(IN), OPTIONAL                  :: nocopy, zero_pad
    TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL :: memory_type

    CHARACTER(len=*), PARAMETER :: routineN = 'ensure_array_size_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: lb_new, lb_orig, &
                                                ub_new, ub_orig, old_size,&
                                                size_increase
    TYPE(dbcsr_memtype_type)                 :: mem_type
    LOGICAL                                  :: dbg, docopy, &
                                                pad
    REAL(kind=real_8), DIMENSION(:), POINTER           :: newarray

!   ---------------------------------------------------------------------------
    !CALL timeset(routineN, error_handler)
    dbg = .FALSE.

    IF (PRESENT(array_resize)) NULLIFY(array_resize)

    IF (PRESENT (nocopy)) THEN
       docopy = .NOT. nocopy
    ELSE
       docopy = .TRUE.
    ENDIF
    IF (PRESENT (memory_type)) THEN
       mem_type = memory_type
    ELSE
       mem_type = dbcsr_memtype_default
    ENDIF
    lb_new = 1
    IF (PRESENT (lb)) lb_new = lb
    pad = .FALSE.
    IF (PRESENT (zero_pad)) pad = zero_pad
    !> Creates a new array if it doesn't yet exist.
    IF (.NOT.ASSOCIATED(array)) THEN
       IF(lb_new /= 1) &
          CPABORT("Arrays must start at 1")
       CALL mem_alloc_d (array, ub, mem_type=mem_type)
       IF (pad .AND. ub .GT. 0) CALL mem_zero_d (array, ub)
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    lb_orig = LBOUND(array,1)
    ub_orig = UBOUND(array,1)
    old_size = ub_orig - lb_orig + 1
    ! The existing array is big enough.
    IF (lb_orig.LE.lb_new .AND. ub_orig.GE.ub) THEN
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    ! A reallocation must be performed
    IF(dbg) WRITE(*,*)routineP//' Current bounds are',lb_orig,':',ub_orig,&
         '; special?' !,mem_type
    !CALL timeset(routineN,timing_handle)
    IF (lb_orig.GT.lb_new) THEN
       IF (PRESENT(factor)) THEN
          size_increase = lb_orig - lb_new
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          lb_new = MIN (lb_orig, lb_new - size_increase)
       ELSE
          lb_new = lb_orig
       ENDIF
    ENDIF
    IF (ub_orig.LT.ub) THEN
       IF (PRESENT(factor)) THEN
          size_increase = ub - ub_orig
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          ub_new = MAX (ub_orig, ub + size_increase)
       ELSE
          ub_new = ub
       ENDIF
    ELSE
       ub_new = ub
    ENDIF
    IF(dbg) WRITE(*,*)routineP//' Resizing to bounds',lb_new,':',ub_new
    !
    ! Deallocates the old array if it's not needed to copy the old data.
    IF(.NOT.docopy) THEN
       IF (PRESENT(array_resize)) THEN
          array_resize => array
          NULLIFY(array)
       ELSE
          CALL mem_dealloc_d (array, mem_type=mem_type)
       ENDIF
    ENDIF
    !
    ! Allocates the new array
    IF(lb_new /= 1) &
       CPABORT("Arrays must start at 1")
    CALL mem_alloc_d (newarray, ub_new-lb_new+1, mem_type)
    !
    ! Now copy and/or zero pad.
    IF(docopy) THEN
       IF(dbg .AND. (lb_new.GT.lb_orig .OR. ub_new.LT.ub_orig))&
            CPABORT("Old extent exceeds the new one.")
       IF (ub_orig-lb_orig+1 .gt. 0) THEN
          !newarray(lb_orig:ub_orig) = array(lb_orig:ub_orig)
          CALL mem_copy_d (newarray(lb_orig:ub_orig),&
               array(lb_orig:ub_orig), ub_orig-lb_orig+1)
       ENDIF
       IF (pad) THEN
          !newarray(lb_new:lb_orig-1) = 0
          CALL mem_zero_d (newarray(lb_new:lb_orig-1), (lb_orig-1)-lb_new+1)
          !newarray(ub_orig+1:ub_new) = 0
          CALL mem_zero_d (newarray(ub_orig+1:ub_new), ub_new-(ub_orig+1)+1)
       ENDIF
       IF (PRESENT(array_resize)) THEN
          array_resize => array
          NULLIFY(array)
       ELSE
          CALL mem_dealloc_d (array, mem_type=mem_type)
       ENDIF
    ELSEIF (pad) THEN
       !newarray(:) = 0.0_real_8
       CALL mem_zero_d (newarray, SIZE(newarray))
    ENDIF
    array => newarray
    IF (dbg) WRITE(*,*)routineP//' New array size', SIZE(array)
    !CALL timestop(error_handler)
  END SUBROUTINE ensure_array_size_d

! **************************************************************************************************
!> \brief Copies memory area
!> \param[out] dst   destination memory
!> \param[in] src    source memory
!> \param[in] n      length of copy
! **************************************************************************************************
  SUBROUTINE mem_copy_d (dst, src, n)
    INTEGER, INTENT(IN) :: n
    REAL(kind=real_8), DIMENSION(1:n), INTENT(OUT) :: dst
    REAL(kind=real_8), DIMENSION(1:n), INTENT(IN) :: src
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst,src)
    dst(:) = src(:)
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_copy_d

! **************************************************************************************************
!> \brief Zeros memory area
!> \param[out] dst   destination memory
!> \param[in] n      length of elements to zero
! **************************************************************************************************
  SUBROUTINE mem_zero_d (dst, n)
    INTEGER, INTENT(IN) :: n
    REAL(kind=real_8), DIMENSION(1:n), INTENT(OUT) :: dst
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst)
    dst(:) = 0.0_real_8
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_zero_d


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] n           length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_d (mem, n, mem_type)
    REAL(kind=real_8), DIMENSION(:), POINTER        :: mem
    INTEGER, INTENT(IN)                   :: n
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. n>1) THEN
       CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi .AND. dbcsr_data_allocation%use_mpi_allocator) THEN
!$omp critical(allocate)
       CALL mp_allocate(mem, n)
!$omp end critical(allocate)
    ELSE
       ALLOCATE(mem(n))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_d


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] sizes length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_d_2d (mem, sizes, mem_type)
    REAL(kind=real_8), DIMENSION(:,:), POINTER      :: mem
    INTEGER, DIMENSION(2), INTENT(IN)     :: sizes
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_d_2d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CPABORT("Accelerator hostalloc not supported for 2D arrays.")
       !CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CPABORT("MPI allocate not supported for 2D arrays.")
       !CALL mp_allocate(mem, n)
    ELSE
       ALLOCATE(mem(sizes(1), sizes(2)))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_d_2d


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_d (mem, mem_type)
    REAL(kind=real_8), DIMENSION(:), POINTER        :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. SIZE(mem)>1) THEN
       CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi .AND. dbcsr_data_allocation%use_mpi_allocator) THEN
       CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_d


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_d_2d (mem, mem_type)
    REAL(kind=real_8), DIMENSION(:,:), POINTER      :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CPABORT("Accelerator host deallocate not supported for 2D arrays.")
       !CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CPABORT("MPI deallocate not supported for 2D arrays.")
       !CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_d_2d


! **************************************************************************************************
!> \brief Sets a rank-2 pointer to rank-1 data using Fortran 2003 pointer
!>        rank remapping.
!> \param r2p ...
!> \param d1 ...
!> \param d2 ...
!> \param r1p ...
! **************************************************************************************************
  SUBROUTINE pointer_d_rank_remap2 (r2p, d1, d2, r1p)
    INTEGER, INTENT(IN)                      :: d1, d2
    REAL(kind=real_8), DIMENSION(:, :), &
      POINTER                                :: r2p
    REAL(kind=real_8), DIMENSION(:), &
      POINTER                                :: r1p

    r2p(1:d1,1:d2) => r1p(1:d1*d2)
  END SUBROUTINE pointer_d_rank_remap2
# 8 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr/data/dbcsr_ptr_util.f90"
! **************************************************************************************************
!> \brief Returns a pointer with different bounds.
!> \param[in] original   original data pointer
!> \param[in] lb lower and upper bound for the new pointer view
!> \param[in] ub lower and upper bound for the new pointer view
!> \return new pointer
! **************************************************************************************************
  FUNCTION pointer_view_s (original, lb, ub) RESULT (view)
    REAL(kind=real_4), DIMENSION(:), POINTER :: original, view
    INTEGER, INTENT(IN)                  :: lb, ub
    view => original(lb:ub)
  END FUNCTION pointer_view_s


! **************************************************************************************************
!> \brief Ensures that an array is appropriately large.
!> \param[in,out] array       array to verify and possibly resize
!> \param[in] lb    (optional) desired array lower bound
!> \param[in] ub    desired array upper bound
!> \param[in] factor          (optional) factor by which to exagerrate
!>                            enlargements
!> \param[in] nocopy          (optional) copy array on enlargement; default
!>                            is to copy
!> \param[in] memory_type     (optional) use special memory
!> \param[in] zero_pad        (optional) zero new allocations; default is to
!>                            write nothing
! **************************************************************************************************
  SUBROUTINE ensure_array_size_s(array, array_resize, lb, ub, factor,&
       nocopy, memory_type, zero_pad)
    REAL(kind=real_4), DIMENSION(:), POINTER                 :: array
    REAL(kind=real_4), DIMENSION(:), POINTER, OPTIONAL       :: array_resize
    INTEGER, INTENT(IN), OPTIONAL                  :: lb
    INTEGER, INTENT(IN)                            :: ub
    REAL(KIND=dp), INTENT(IN), OPTIONAL            :: factor
    LOGICAL, INTENT(IN), OPTIONAL                  :: nocopy, zero_pad
    TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL :: memory_type

    CHARACTER(len=*), PARAMETER :: routineN = 'ensure_array_size_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: lb_new, lb_orig, &
                                                ub_new, ub_orig, old_size,&
                                                size_increase
    TYPE(dbcsr_memtype_type)                 :: mem_type
    LOGICAL                                  :: dbg, docopy, &
                                                pad
    REAL(kind=real_4), DIMENSION(:), POINTER           :: newarray

!   ---------------------------------------------------------------------------
    !CALL timeset(routineN, error_handler)
    dbg = .FALSE.

    IF (PRESENT(array_resize)) NULLIFY(array_resize)

    IF (PRESENT (nocopy)) THEN
       docopy = .NOT. nocopy
    ELSE
       docopy = .TRUE.
    ENDIF
    IF (PRESENT (memory_type)) THEN
       mem_type = memory_type
    ELSE
       mem_type = dbcsr_memtype_default
    ENDIF
    lb_new = 1
    IF (PRESENT (lb)) lb_new = lb
    pad = .FALSE.
    IF (PRESENT (zero_pad)) pad = zero_pad
    !> Creates a new array if it doesn't yet exist.
    IF (.NOT.ASSOCIATED(array)) THEN
       IF(lb_new /= 1) &
          CPABORT("Arrays must start at 1")
       CALL mem_alloc_s (array, ub, mem_type=mem_type)
       IF (pad .AND. ub .GT. 0) CALL mem_zero_s (array, ub)
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    lb_orig = LBOUND(array,1)
    ub_orig = UBOUND(array,1)
    old_size = ub_orig - lb_orig + 1
    ! The existing array is big enough.
    IF (lb_orig.LE.lb_new .AND. ub_orig.GE.ub) THEN
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    ! A reallocation must be performed
    IF(dbg) WRITE(*,*)routineP//' Current bounds are',lb_orig,':',ub_orig,&
         '; special?' !,mem_type
    !CALL timeset(routineN,timing_handle)
    IF (lb_orig.GT.lb_new) THEN
       IF (PRESENT(factor)) THEN
          size_increase = lb_orig - lb_new
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          lb_new = MIN (lb_orig, lb_new - size_increase)
       ELSE
          lb_new = lb_orig
       ENDIF
    ENDIF
    IF (ub_orig.LT.ub) THEN
       IF (PRESENT(factor)) THEN
          size_increase = ub - ub_orig
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          ub_new = MAX (ub_orig, ub + size_increase)
       ELSE
          ub_new = ub
       ENDIF
    ELSE
       ub_new = ub
    ENDIF
    IF(dbg) WRITE(*,*)routineP//' Resizing to bounds',lb_new,':',ub_new
    !
    ! Deallocates the old array if it's not needed to copy the old data.
    IF(.NOT.docopy) THEN
       IF (PRESENT(array_resize)) THEN
          array_resize => array
          NULLIFY(array)
       ELSE
          CALL mem_dealloc_s (array, mem_type=mem_type)
       ENDIF
    ENDIF
    !
    ! Allocates the new array
    IF(lb_new /= 1) &
       CPABORT("Arrays must start at 1")
    CALL mem_alloc_s (newarray, ub_new-lb_new+1, mem_type)
    !
    ! Now copy and/or zero pad.
    IF(docopy) THEN
       IF(dbg .AND. (lb_new.GT.lb_orig .OR. ub_new.LT.ub_orig))&
            CPABORT("Old extent exceeds the new one.")
       IF (ub_orig-lb_orig+1 .gt. 0) THEN
          !newarray(lb_orig:ub_orig) = array(lb_orig:ub_orig)
          CALL mem_copy_s (newarray(lb_orig:ub_orig),&
               array(lb_orig:ub_orig), ub_orig-lb_orig+1)
       ENDIF
       IF (pad) THEN
          !newarray(lb_new:lb_orig-1) = 0
          CALL mem_zero_s (newarray(lb_new:lb_orig-1), (lb_orig-1)-lb_new+1)
          !newarray(ub_orig+1:ub_new) = 0
          CALL mem_zero_s (newarray(ub_orig+1:ub_new), ub_new-(ub_orig+1)+1)
       ENDIF
       IF (PRESENT(array_resize)) THEN
          array_resize => array
          NULLIFY(array)
       ELSE
          CALL mem_dealloc_s (array, mem_type=mem_type)
       ENDIF
    ELSEIF (pad) THEN
       !newarray(:) = 0.0_real_4
       CALL mem_zero_s (newarray, SIZE(newarray))
    ENDIF
    array => newarray
    IF (dbg) WRITE(*,*)routineP//' New array size', SIZE(array)
    !CALL timestop(error_handler)
  END SUBROUTINE ensure_array_size_s

! **************************************************************************************************
!> \brief Copies memory area
!> \param[out] dst   destination memory
!> \param[in] src    source memory
!> \param[in] n      length of copy
! **************************************************************************************************
  SUBROUTINE mem_copy_s (dst, src, n)
    INTEGER, INTENT(IN) :: n
    REAL(kind=real_4), DIMENSION(1:n), INTENT(OUT) :: dst
    REAL(kind=real_4), DIMENSION(1:n), INTENT(IN) :: src
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst,src)
    dst(:) = src(:)
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_copy_s

! **************************************************************************************************
!> \brief Zeros memory area
!> \param[out] dst   destination memory
!> \param[in] n      length of elements to zero
! **************************************************************************************************
  SUBROUTINE mem_zero_s (dst, n)
    INTEGER, INTENT(IN) :: n
    REAL(kind=real_4), DIMENSION(1:n), INTENT(OUT) :: dst
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst)
    dst(:) = 0.0_real_4
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_zero_s


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] n           length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_s (mem, n, mem_type)
    REAL(kind=real_4), DIMENSION(:), POINTER        :: mem
    INTEGER, INTENT(IN)                   :: n
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_s', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. n>1) THEN
       CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi .AND. dbcsr_data_allocation%use_mpi_allocator) THEN
!$omp critical(allocate)
       CALL mp_allocate(mem, n)
!$omp end critical(allocate)
    ELSE
       ALLOCATE(mem(n))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_s


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] sizes length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_s_2d (mem, sizes, mem_type)
    REAL(kind=real_4), DIMENSION(:,:), POINTER      :: mem
    INTEGER, DIMENSION(2), INTENT(IN)     :: sizes
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_s_2d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CPABORT("Accelerator hostalloc not supported for 2D arrays.")
       !CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CPABORT("MPI allocate not supported for 2D arrays.")
       !CALL mp_allocate(mem, n)
    ELSE
       ALLOCATE(mem(sizes(1), sizes(2)))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_s_2d


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_s (mem, mem_type)
    REAL(kind=real_4), DIMENSION(:), POINTER        :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_s', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. SIZE(mem)>1) THEN
       CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi .AND. dbcsr_data_allocation%use_mpi_allocator) THEN
       CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_s


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_s_2d (mem, mem_type)
    REAL(kind=real_4), DIMENSION(:,:), POINTER      :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_s', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CPABORT("Accelerator host deallocate not supported for 2D arrays.")
       !CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CPABORT("MPI deallocate not supported for 2D arrays.")
       !CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_s_2d


! **************************************************************************************************
!> \brief Sets a rank-2 pointer to rank-1 data using Fortran 2003 pointer
!>        rank remapping.
!> \param r2p ...
!> \param d1 ...
!> \param d2 ...
!> \param r1p ...
! **************************************************************************************************
  SUBROUTINE pointer_s_rank_remap2 (r2p, d1, d2, r1p)
    INTEGER, INTENT(IN)                      :: d1, d2
    REAL(kind=real_4), DIMENSION(:, :), &
      POINTER                                :: r2p
    REAL(kind=real_4), DIMENSION(:), &
      POINTER                                :: r1p

    r2p(1:d1,1:d2) => r1p(1:d1*d2)
  END SUBROUTINE pointer_s_rank_remap2
# 8 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr/data/dbcsr_ptr_util.f90"
! **************************************************************************************************
!> \brief Returns a pointer with different bounds.
!> \param[in] original   original data pointer
!> \param[in] lb lower and upper bound for the new pointer view
!> \param[in] ub lower and upper bound for the new pointer view
!> \return new pointer
! **************************************************************************************************
  FUNCTION pointer_view_z (original, lb, ub) RESULT (view)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER :: original, view
    INTEGER, INTENT(IN)                  :: lb, ub
    view => original(lb:ub)
  END FUNCTION pointer_view_z


! **************************************************************************************************
!> \brief Ensures that an array is appropriately large.
!> \param[in,out] array       array to verify and possibly resize
!> \param[in] lb    (optional) desired array lower bound
!> \param[in] ub    desired array upper bound
!> \param[in] factor          (optional) factor by which to exagerrate
!>                            enlargements
!> \param[in] nocopy          (optional) copy array on enlargement; default
!>                            is to copy
!> \param[in] memory_type     (optional) use special memory
!> \param[in] zero_pad        (optional) zero new allocations; default is to
!>                            write nothing
! **************************************************************************************************
  SUBROUTINE ensure_array_size_z(array, array_resize, lb, ub, factor,&
       nocopy, memory_type, zero_pad)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER                 :: array
    COMPLEX(kind=real_8), DIMENSION(:), POINTER, OPTIONAL       :: array_resize
    INTEGER, INTENT(IN), OPTIONAL                  :: lb
    INTEGER, INTENT(IN)                            :: ub
    REAL(KIND=dp), INTENT(IN), OPTIONAL            :: factor
    LOGICAL, INTENT(IN), OPTIONAL                  :: nocopy, zero_pad
    TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL :: memory_type

    CHARACTER(len=*), PARAMETER :: routineN = 'ensure_array_size_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: lb_new, lb_orig, &
                                                ub_new, ub_orig, old_size,&
                                                size_increase
    TYPE(dbcsr_memtype_type)                 :: mem_type
    LOGICAL                                  :: dbg, docopy, &
                                                pad
    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: newarray

!   ---------------------------------------------------------------------------
    !CALL timeset(routineN, error_handler)
    dbg = .FALSE.

    IF (PRESENT(array_resize)) NULLIFY(array_resize)

    IF (PRESENT (nocopy)) THEN
       docopy = .NOT. nocopy
    ELSE
       docopy = .TRUE.
    ENDIF
    IF (PRESENT (memory_type)) THEN
       mem_type = memory_type
    ELSE
       mem_type = dbcsr_memtype_default
    ENDIF
    lb_new = 1
    IF (PRESENT (lb)) lb_new = lb
    pad = .FALSE.
    IF (PRESENT (zero_pad)) pad = zero_pad
    !> Creates a new array if it doesn't yet exist.
    IF (.NOT.ASSOCIATED(array)) THEN
       IF(lb_new /= 1) &
          CPABORT("Arrays must start at 1")
       CALL mem_alloc_z (array, ub, mem_type=mem_type)
       IF (pad .AND. ub .GT. 0) CALL mem_zero_z (array, ub)
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    lb_orig = LBOUND(array,1)
    ub_orig = UBOUND(array,1)
    old_size = ub_orig - lb_orig + 1
    ! The existing array is big enough.
    IF (lb_orig.LE.lb_new .AND. ub_orig.GE.ub) THEN
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    ! A reallocation must be performed
    IF(dbg) WRITE(*,*)routineP//' Current bounds are',lb_orig,':',ub_orig,&
         '; special?' !,mem_type
    !CALL timeset(routineN,timing_handle)
    IF (lb_orig.GT.lb_new) THEN
       IF (PRESENT(factor)) THEN
          size_increase = lb_orig - lb_new
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          lb_new = MIN (lb_orig, lb_new - size_increase)
       ELSE
          lb_new = lb_orig
       ENDIF
    ENDIF
    IF (ub_orig.LT.ub) THEN
       IF (PRESENT(factor)) THEN
          size_increase = ub - ub_orig
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          ub_new = MAX (ub_orig, ub + size_increase)
       ELSE
          ub_new = ub
       ENDIF
    ELSE
       ub_new = ub
    ENDIF
    IF(dbg) WRITE(*,*)routineP//' Resizing to bounds',lb_new,':',ub_new
    !
    ! Deallocates the old array if it's not needed to copy the old data.
    IF(.NOT.docopy) THEN
       IF (PRESENT(array_resize)) THEN
          array_resize => array
          NULLIFY(array)
       ELSE
          CALL mem_dealloc_z (array, mem_type=mem_type)
       ENDIF
    ENDIF
    !
    ! Allocates the new array
    IF(lb_new /= 1) &
       CPABORT("Arrays must start at 1")
    CALL mem_alloc_z (newarray, ub_new-lb_new+1, mem_type)
    !
    ! Now copy and/or zero pad.
    IF(docopy) THEN
       IF(dbg .AND. (lb_new.GT.lb_orig .OR. ub_new.LT.ub_orig))&
            CPABORT("Old extent exceeds the new one.")
       IF (ub_orig-lb_orig+1 .gt. 0) THEN
          !newarray(lb_orig:ub_orig) = array(lb_orig:ub_orig)
          CALL mem_copy_z (newarray(lb_orig:ub_orig),&
               array(lb_orig:ub_orig), ub_orig-lb_orig+1)
       ENDIF
       IF (pad) THEN
          !newarray(lb_new:lb_orig-1) = 0
          CALL mem_zero_z (newarray(lb_new:lb_orig-1), (lb_orig-1)-lb_new+1)
          !newarray(ub_orig+1:ub_new) = 0
          CALL mem_zero_z (newarray(ub_orig+1:ub_new), ub_new-(ub_orig+1)+1)
       ENDIF
       IF (PRESENT(array_resize)) THEN
          array_resize => array
          NULLIFY(array)
       ELSE
          CALL mem_dealloc_z (array, mem_type=mem_type)
       ENDIF
    ELSEIF (pad) THEN
       !newarray(:) = CMPLX(0.0, 0.0, real_8)
       CALL mem_zero_z (newarray, SIZE(newarray))
    ENDIF
    array => newarray
    IF (dbg) WRITE(*,*)routineP//' New array size', SIZE(array)
    !CALL timestop(error_handler)
  END SUBROUTINE ensure_array_size_z

! **************************************************************************************************
!> \brief Copies memory area
!> \param[out] dst   destination memory
!> \param[in] src    source memory
!> \param[in] n      length of copy
! **************************************************************************************************
  SUBROUTINE mem_copy_z (dst, src, n)
    INTEGER, INTENT(IN) :: n
    COMPLEX(kind=real_8), DIMENSION(1:n), INTENT(OUT) :: dst
    COMPLEX(kind=real_8), DIMENSION(1:n), INTENT(IN) :: src
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst,src)
    dst(:) = src(:)
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_copy_z

! **************************************************************************************************
!> \brief Zeros memory area
!> \param[out] dst   destination memory
!> \param[in] n      length of elements to zero
! **************************************************************************************************
  SUBROUTINE mem_zero_z (dst, n)
    INTEGER, INTENT(IN) :: n
    COMPLEX(kind=real_8), DIMENSION(1:n), INTENT(OUT) :: dst
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst)
    dst(:) = CMPLX(0.0, 0.0, real_8)
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_zero_z


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] n           length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_z (mem, n, mem_type)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER        :: mem
    INTEGER, INTENT(IN)                   :: n
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_z', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. n>1) THEN
       CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi .AND. dbcsr_data_allocation%use_mpi_allocator) THEN
!$omp critical(allocate)
       CALL mp_allocate(mem, n)
!$omp end critical(allocate)
    ELSE
       ALLOCATE(mem(n))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_z


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] sizes length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_z_2d (mem, sizes, mem_type)
    COMPLEX(kind=real_8), DIMENSION(:,:), POINTER      :: mem
    INTEGER, DIMENSION(2), INTENT(IN)     :: sizes
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_z_2d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CPABORT("Accelerator hostalloc not supported for 2D arrays.")
       !CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CPABORT("MPI allocate not supported for 2D arrays.")
       !CALL mp_allocate(mem, n)
    ELSE
       ALLOCATE(mem(sizes(1), sizes(2)))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_z_2d


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_z (mem, mem_type)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER        :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_z', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. SIZE(mem)>1) THEN
       CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi .AND. dbcsr_data_allocation%use_mpi_allocator) THEN
       CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_z


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_z_2d (mem, mem_type)
    COMPLEX(kind=real_8), DIMENSION(:,:), POINTER      :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_z', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CPABORT("Accelerator host deallocate not supported for 2D arrays.")
       !CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CPABORT("MPI deallocate not supported for 2D arrays.")
       !CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_z_2d


! **************************************************************************************************
!> \brief Sets a rank-2 pointer to rank-1 data using Fortran 2003 pointer
!>        rank remapping.
!> \param r2p ...
!> \param d1 ...
!> \param d2 ...
!> \param r1p ...
! **************************************************************************************************
  SUBROUTINE pointer_z_rank_remap2 (r2p, d1, d2, r1p)
    INTEGER, INTENT(IN)                      :: d1, d2
    COMPLEX(kind=real_8), DIMENSION(:, :), &
      POINTER                                :: r2p
    COMPLEX(kind=real_8), DIMENSION(:), &
      POINTER                                :: r1p

    r2p(1:d1,1:d2) => r1p(1:d1*d2)
  END SUBROUTINE pointer_z_rank_remap2
# 8 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr/data/dbcsr_ptr_util.f90"
! **************************************************************************************************
!> \brief Returns a pointer with different bounds.
!> \param[in] original   original data pointer
!> \param[in] lb lower and upper bound for the new pointer view
!> \param[in] ub lower and upper bound for the new pointer view
!> \return new pointer
! **************************************************************************************************
  FUNCTION pointer_view_c (original, lb, ub) RESULT (view)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER :: original, view
    INTEGER, INTENT(IN)                  :: lb, ub
    view => original(lb:ub)
  END FUNCTION pointer_view_c


! **************************************************************************************************
!> \brief Ensures that an array is appropriately large.
!> \param[in,out] array       array to verify and possibly resize
!> \param[in] lb    (optional) desired array lower bound
!> \param[in] ub    desired array upper bound
!> \param[in] factor          (optional) factor by which to exagerrate
!>                            enlargements
!> \param[in] nocopy          (optional) copy array on enlargement; default
!>                            is to copy
!> \param[in] memory_type     (optional) use special memory
!> \param[in] zero_pad        (optional) zero new allocations; default is to
!>                            write nothing
! **************************************************************************************************
  SUBROUTINE ensure_array_size_c(array, array_resize, lb, ub, factor,&
       nocopy, memory_type, zero_pad)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER                 :: array
    COMPLEX(kind=real_4), DIMENSION(:), POINTER, OPTIONAL       :: array_resize
    INTEGER, INTENT(IN), OPTIONAL                  :: lb
    INTEGER, INTENT(IN)                            :: ub
    REAL(KIND=dp), INTENT(IN), OPTIONAL            :: factor
    LOGICAL, INTENT(IN), OPTIONAL                  :: nocopy, zero_pad
    TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL :: memory_type

    CHARACTER(len=*), PARAMETER :: routineN = 'ensure_array_size_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: lb_new, lb_orig, &
                                                ub_new, ub_orig, old_size,&
                                                size_increase
    TYPE(dbcsr_memtype_type)                 :: mem_type
    LOGICAL                                  :: dbg, docopy, &
                                                pad
    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: newarray

!   ---------------------------------------------------------------------------
    !CALL timeset(routineN, error_handler)
    dbg = .FALSE.

    IF (PRESENT(array_resize)) NULLIFY(array_resize)

    IF (PRESENT (nocopy)) THEN
       docopy = .NOT. nocopy
    ELSE
       docopy = .TRUE.
    ENDIF
    IF (PRESENT (memory_type)) THEN
       mem_type = memory_type
    ELSE
       mem_type = dbcsr_memtype_default
    ENDIF
    lb_new = 1
    IF (PRESENT (lb)) lb_new = lb
    pad = .FALSE.
    IF (PRESENT (zero_pad)) pad = zero_pad
    !> Creates a new array if it doesn't yet exist.
    IF (.NOT.ASSOCIATED(array)) THEN
       IF(lb_new /= 1) &
          CPABORT("Arrays must start at 1")
       CALL mem_alloc_c (array, ub, mem_type=mem_type)
       IF (pad .AND. ub .GT. 0) CALL mem_zero_c (array, ub)
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    lb_orig = LBOUND(array,1)
    ub_orig = UBOUND(array,1)
    old_size = ub_orig - lb_orig + 1
    ! The existing array is big enough.
    IF (lb_orig.LE.lb_new .AND. ub_orig.GE.ub) THEN
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    ! A reallocation must be performed
    IF(dbg) WRITE(*,*)routineP//' Current bounds are',lb_orig,':',ub_orig,&
         '; special?' !,mem_type
    !CALL timeset(routineN,timing_handle)
    IF (lb_orig.GT.lb_new) THEN
       IF (PRESENT(factor)) THEN
          size_increase = lb_orig - lb_new
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          lb_new = MIN (lb_orig, lb_new - size_increase)
       ELSE
          lb_new = lb_orig
       ENDIF
    ENDIF
    IF (ub_orig.LT.ub) THEN
       IF (PRESENT(factor)) THEN
          size_increase = ub - ub_orig
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          ub_new = MAX (ub_orig, ub + size_increase)
       ELSE
          ub_new = ub
       ENDIF
    ELSE
       ub_new = ub
    ENDIF
    IF(dbg) WRITE(*,*)routineP//' Resizing to bounds',lb_new,':',ub_new
    !
    ! Deallocates the old array if it's not needed to copy the old data.
    IF(.NOT.docopy) THEN
       IF (PRESENT(array_resize)) THEN
          array_resize => array
          NULLIFY(array)
       ELSE
          CALL mem_dealloc_c (array, mem_type=mem_type)
       ENDIF
    ENDIF
    !
    ! Allocates the new array
    IF(lb_new /= 1) &
       CPABORT("Arrays must start at 1")
    CALL mem_alloc_c (newarray, ub_new-lb_new+1, mem_type)
    !
    ! Now copy and/or zero pad.
    IF(docopy) THEN
       IF(dbg .AND. (lb_new.GT.lb_orig .OR. ub_new.LT.ub_orig))&
            CPABORT("Old extent exceeds the new one.")
       IF (ub_orig-lb_orig+1 .gt. 0) THEN
          !newarray(lb_orig:ub_orig) = array(lb_orig:ub_orig)
          CALL mem_copy_c (newarray(lb_orig:ub_orig),&
               array(lb_orig:ub_orig), ub_orig-lb_orig+1)
       ENDIF
       IF (pad) THEN
          !newarray(lb_new:lb_orig-1) = 0
          CALL mem_zero_c (newarray(lb_new:lb_orig-1), (lb_orig-1)-lb_new+1)
          !newarray(ub_orig+1:ub_new) = 0
          CALL mem_zero_c (newarray(ub_orig+1:ub_new), ub_new-(ub_orig+1)+1)
       ENDIF
       IF (PRESENT(array_resize)) THEN
          array_resize => array
          NULLIFY(array)
       ELSE
          CALL mem_dealloc_c (array, mem_type=mem_type)
       ENDIF
    ELSEIF (pad) THEN
       !newarray(:) = CMPLX(0.0, 0.0, real_4)
       CALL mem_zero_c (newarray, SIZE(newarray))
    ENDIF
    array => newarray
    IF (dbg) WRITE(*,*)routineP//' New array size', SIZE(array)
    !CALL timestop(error_handler)
  END SUBROUTINE ensure_array_size_c

! **************************************************************************************************
!> \brief Copies memory area
!> \param[out] dst   destination memory
!> \param[in] src    source memory
!> \param[in] n      length of copy
! **************************************************************************************************
  SUBROUTINE mem_copy_c (dst, src, n)
    INTEGER, INTENT(IN) :: n
    COMPLEX(kind=real_4), DIMENSION(1:n), INTENT(OUT) :: dst
    COMPLEX(kind=real_4), DIMENSION(1:n), INTENT(IN) :: src
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst,src)
    dst(:) = src(:)
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_copy_c

! **************************************************************************************************
!> \brief Zeros memory area
!> \param[out] dst   destination memory
!> \param[in] n      length of elements to zero
! **************************************************************************************************
  SUBROUTINE mem_zero_c (dst, n)
    INTEGER, INTENT(IN) :: n
    COMPLEX(kind=real_4), DIMENSION(1:n), INTENT(OUT) :: dst
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst)
    dst(:) = CMPLX(0.0, 0.0, real_4)
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_zero_c


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] n           length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_c (mem, n, mem_type)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER        :: mem
    INTEGER, INTENT(IN)                   :: n
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_c', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. n>1) THEN
       CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi .AND. dbcsr_data_allocation%use_mpi_allocator) THEN
!$omp critical(allocate)
       CALL mp_allocate(mem, n)
!$omp end critical(allocate)
    ELSE
       ALLOCATE(mem(n))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_c


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] sizes length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_c_2d (mem, sizes, mem_type)
    COMPLEX(kind=real_4), DIMENSION(:,:), POINTER      :: mem
    INTEGER, DIMENSION(2), INTENT(IN)     :: sizes
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_c_2d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CPABORT("Accelerator hostalloc not supported for 2D arrays.")
       !CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CPABORT("MPI allocate not supported for 2D arrays.")
       !CALL mp_allocate(mem, n)
    ELSE
       ALLOCATE(mem(sizes(1), sizes(2)))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_c_2d


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_c (mem, mem_type)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER        :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_c', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. SIZE(mem)>1) THEN
       CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi .AND. dbcsr_data_allocation%use_mpi_allocator) THEN
       CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_c


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_c_2d (mem, mem_type)
    COMPLEX(kind=real_4), DIMENSION(:,:), POINTER      :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_c', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CPABORT("Accelerator host deallocate not supported for 2D arrays.")
       !CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CPABORT("MPI deallocate not supported for 2D arrays.")
       !CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_c_2d


! **************************************************************************************************
!> \brief Sets a rank-2 pointer to rank-1 data using Fortran 2003 pointer
!>        rank remapping.
!> \param r2p ...
!> \param d1 ...
!> \param d2 ...
!> \param r1p ...
! **************************************************************************************************
  SUBROUTINE pointer_c_rank_remap2 (r2p, d1, d2, r1p)
    INTEGER, INTENT(IN)                      :: d1, d2
    COMPLEX(kind=real_4), DIMENSION(:, :), &
      POINTER                                :: r2p
    COMPLEX(kind=real_4), DIMENSION(:), &
      POINTER                                :: r1p

    r2p(1:d1,1:d2) => r1p(1:d1*d2)
  END SUBROUTINE pointer_c_rank_remap2
# 8 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr/data/dbcsr_ptr_util.f90"
! **************************************************************************************************
!> \brief Returns a pointer with different bounds.
!> \param[in] original   original data pointer
!> \param[in] lb lower and upper bound for the new pointer view
!> \param[in] ub lower and upper bound for the new pointer view
!> \return new pointer
! **************************************************************************************************
  FUNCTION pointer_view_i (original, lb, ub) RESULT (view)
    INTEGER(kind=int_4), DIMENSION(:), POINTER :: original, view
    INTEGER, INTENT(IN)                  :: lb, ub
    view => original(lb:ub)
  END FUNCTION pointer_view_i


! **************************************************************************************************
!> \brief Ensures that an array is appropriately large.
!> \param[in,out] array       array to verify and possibly resize
!> \param[in] lb    (optional) desired array lower bound
!> \param[in] ub    desired array upper bound
!> \param[in] factor          (optional) factor by which to exagerrate
!>                            enlargements
!> \param[in] nocopy          (optional) copy array on enlargement; default
!>                            is to copy
!> \param[in] memory_type     (optional) use special memory
!> \param[in] zero_pad        (optional) zero new allocations; default is to
!>                            write nothing
! **************************************************************************************************
  SUBROUTINE ensure_array_size_i(array, array_resize, lb, ub, factor,&
       nocopy, memory_type, zero_pad)
    INTEGER(kind=int_4), DIMENSION(:), POINTER                 :: array
    INTEGER(kind=int_4), DIMENSION(:), POINTER, OPTIONAL       :: array_resize
    INTEGER, INTENT(IN), OPTIONAL                  :: lb
    INTEGER, INTENT(IN)                            :: ub
    REAL(KIND=dp), INTENT(IN), OPTIONAL            :: factor
    LOGICAL, INTENT(IN), OPTIONAL                  :: nocopy, zero_pad
    TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL :: memory_type

    CHARACTER(len=*), PARAMETER :: routineN = 'ensure_array_size_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: lb_new, lb_orig, &
                                                ub_new, ub_orig, old_size,&
                                                size_increase
    TYPE(dbcsr_memtype_type)                 :: mem_type
    LOGICAL                                  :: dbg, docopy, &
                                                pad
    INTEGER(kind=int_4), DIMENSION(:), POINTER           :: newarray

!   ---------------------------------------------------------------------------
    !CALL timeset(routineN, error_handler)
    dbg = .FALSE.

    IF (PRESENT(array_resize)) NULLIFY(array_resize)

    IF (PRESENT (nocopy)) THEN
       docopy = .NOT. nocopy
    ELSE
       docopy = .TRUE.
    ENDIF
    IF (PRESENT (memory_type)) THEN
       mem_type = memory_type
    ELSE
       mem_type = dbcsr_memtype_default
    ENDIF
    lb_new = 1
    IF (PRESENT (lb)) lb_new = lb
    pad = .FALSE.
    IF (PRESENT (zero_pad)) pad = zero_pad
    !> Creates a new array if it doesn't yet exist.
    IF (.NOT.ASSOCIATED(array)) THEN
       IF(lb_new /= 1) &
          CPABORT("Arrays must start at 1")
       CALL mem_alloc_i (array, ub, mem_type=mem_type)
       IF (pad .AND. ub .GT. 0) CALL mem_zero_i (array, ub)
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    lb_orig = LBOUND(array,1)
    ub_orig = UBOUND(array,1)
    old_size = ub_orig - lb_orig + 1
    ! The existing array is big enough.
    IF (lb_orig.LE.lb_new .AND. ub_orig.GE.ub) THEN
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    ! A reallocation must be performed
    IF(dbg) WRITE(*,*)routineP//' Current bounds are',lb_orig,':',ub_orig,&
         '; special?' !,mem_type
    !CALL timeset(routineN,timing_handle)
    IF (lb_orig.GT.lb_new) THEN
       IF (PRESENT(factor)) THEN
          size_increase = lb_orig - lb_new
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          lb_new = MIN (lb_orig, lb_new - size_increase)
       ELSE
          lb_new = lb_orig
       ENDIF
    ENDIF
    IF (ub_orig.LT.ub) THEN
       IF (PRESENT(factor)) THEN
          size_increase = ub - ub_orig
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          ub_new = MAX (ub_orig, ub + size_increase)
       ELSE
          ub_new = ub
       ENDIF
    ELSE
       ub_new = ub
    ENDIF
    IF(dbg) WRITE(*,*)routineP//' Resizing to bounds',lb_new,':',ub_new
    !
    ! Deallocates the old array if it's not needed to copy the old data.
    IF(.NOT.docopy) THEN
       IF (PRESENT(array_resize)) THEN
          array_resize => array
          NULLIFY(array)
       ELSE
          CALL mem_dealloc_i (array, mem_type=mem_type)
       ENDIF
    ENDIF
    !
    ! Allocates the new array
    IF(lb_new /= 1) &
       CPABORT("Arrays must start at 1")
    CALL mem_alloc_i (newarray, ub_new-lb_new+1, mem_type)
    !
    ! Now copy and/or zero pad.
    IF(docopy) THEN
       IF(dbg .AND. (lb_new.GT.lb_orig .OR. ub_new.LT.ub_orig))&
            CPABORT("Old extent exceeds the new one.")
       IF (ub_orig-lb_orig+1 .gt. 0) THEN
          !newarray(lb_orig:ub_orig) = array(lb_orig:ub_orig)
          CALL mem_copy_i (newarray(lb_orig:ub_orig),&
               array(lb_orig:ub_orig), ub_orig-lb_orig+1)
       ENDIF
       IF (pad) THEN
          !newarray(lb_new:lb_orig-1) = 0
          CALL mem_zero_i (newarray(lb_new:lb_orig-1), (lb_orig-1)-lb_new+1)
          !newarray(ub_orig+1:ub_new) = 0
          CALL mem_zero_i (newarray(ub_orig+1:ub_new), ub_new-(ub_orig+1)+1)
       ENDIF
       IF (PRESENT(array_resize)) THEN
          array_resize => array
          NULLIFY(array)
       ELSE
          CALL mem_dealloc_i (array, mem_type=mem_type)
       ENDIF
    ELSEIF (pad) THEN
       !newarray(:) = 0
       CALL mem_zero_i (newarray, SIZE(newarray))
    ENDIF
    array => newarray
    IF (dbg) WRITE(*,*)routineP//' New array size', SIZE(array)
    !CALL timestop(error_handler)
  END SUBROUTINE ensure_array_size_i

! **************************************************************************************************
!> \brief Copies memory area
!> \param[out] dst   destination memory
!> \param[in] src    source memory
!> \param[in] n      length of copy
! **************************************************************************************************
  SUBROUTINE mem_copy_i (dst, src, n)
    INTEGER, INTENT(IN) :: n
    INTEGER(kind=int_4), DIMENSION(1:n), INTENT(OUT) :: dst
    INTEGER(kind=int_4), DIMENSION(1:n), INTENT(IN) :: src
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst,src)
    dst(:) = src(:)
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_copy_i

! **************************************************************************************************
!> \brief Zeros memory area
!> \param[out] dst   destination memory
!> \param[in] n      length of elements to zero
! **************************************************************************************************
  SUBROUTINE mem_zero_i (dst, n)
    INTEGER, INTENT(IN) :: n
    INTEGER(kind=int_4), DIMENSION(1:n), INTENT(OUT) :: dst
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst)
    dst(:) = 0
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_zero_i


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] n           length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_i (mem, n, mem_type)
    INTEGER(kind=int_4), DIMENSION(:), POINTER        :: mem
    INTEGER, INTENT(IN)                   :: n
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_i', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. n>1) THEN
       CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi .AND. dbcsr_data_allocation%use_mpi_allocator) THEN
!$omp critical(allocate)
       CALL mp_allocate(mem, n)
!$omp end critical(allocate)
    ELSE
       ALLOCATE(mem(n))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_i


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] sizes length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_i_2d (mem, sizes, mem_type)
    INTEGER(kind=int_4), DIMENSION(:,:), POINTER      :: mem
    INTEGER, DIMENSION(2), INTENT(IN)     :: sizes
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_i_2d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CPABORT("Accelerator hostalloc not supported for 2D arrays.")
       !CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CPABORT("MPI allocate not supported for 2D arrays.")
       !CALL mp_allocate(mem, n)
    ELSE
       ALLOCATE(mem(sizes(1), sizes(2)))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_i_2d


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_i (mem, mem_type)
    INTEGER(kind=int_4), DIMENSION(:), POINTER        :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_i', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. SIZE(mem)>1) THEN
       CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi .AND. dbcsr_data_allocation%use_mpi_allocator) THEN
       CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_i


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_i_2d (mem, mem_type)
    INTEGER(kind=int_4), DIMENSION(:,:), POINTER      :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_i', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CPABORT("Accelerator host deallocate not supported for 2D arrays.")
       !CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CPABORT("MPI deallocate not supported for 2D arrays.")
       !CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_i_2d


! **************************************************************************************************
!> \brief Sets a rank-2 pointer to rank-1 data using Fortran 2003 pointer
!>        rank remapping.
!> \param r2p ...
!> \param d1 ...
!> \param d2 ...
!> \param r1p ...
! **************************************************************************************************
  SUBROUTINE pointer_i_rank_remap2 (r2p, d1, d2, r1p)
    INTEGER, INTENT(IN)                      :: d1, d2
    INTEGER(kind=int_4), DIMENSION(:, :), &
      POINTER                                :: r2p
    INTEGER(kind=int_4), DIMENSION(:), &
      POINTER                                :: r1p

    r2p(1:d1,1:d2) => r1p(1:d1*d2)
  END SUBROUTINE pointer_i_rank_remap2
# 8 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr/data/dbcsr_ptr_util.f90"
! **************************************************************************************************
!> \brief Returns a pointer with different bounds.
!> \param[in] original   original data pointer
!> \param[in] lb lower and upper bound for the new pointer view
!> \param[in] ub lower and upper bound for the new pointer view
!> \return new pointer
! **************************************************************************************************
  FUNCTION pointer_view_l (original, lb, ub) RESULT (view)
    INTEGER(kind=int_8), DIMENSION(:), POINTER :: original, view
    INTEGER, INTENT(IN)                  :: lb, ub
    view => original(lb:ub)
  END FUNCTION pointer_view_l


! **************************************************************************************************
!> \brief Ensures that an array is appropriately large.
!> \param[in,out] array       array to verify and possibly resize
!> \param[in] lb    (optional) desired array lower bound
!> \param[in] ub    desired array upper bound
!> \param[in] factor          (optional) factor by which to exagerrate
!>                            enlargements
!> \param[in] nocopy          (optional) copy array on enlargement; default
!>                            is to copy
!> \param[in] memory_type     (optional) use special memory
!> \param[in] zero_pad        (optional) zero new allocations; default is to
!>                            write nothing
! **************************************************************************************************
  SUBROUTINE ensure_array_size_l(array, array_resize, lb, ub, factor,&
       nocopy, memory_type, zero_pad)
    INTEGER(kind=int_8), DIMENSION(:), POINTER                 :: array
    INTEGER(kind=int_8), DIMENSION(:), POINTER, OPTIONAL       :: array_resize
    INTEGER, INTENT(IN), OPTIONAL                  :: lb
    INTEGER, INTENT(IN)                            :: ub
    REAL(KIND=dp), INTENT(IN), OPTIONAL            :: factor
    LOGICAL, INTENT(IN), OPTIONAL                  :: nocopy, zero_pad
    TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL :: memory_type

    CHARACTER(len=*), PARAMETER :: routineN = 'ensure_array_size_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: lb_new, lb_orig, &
                                                ub_new, ub_orig, old_size,&
                                                size_increase
    TYPE(dbcsr_memtype_type)                 :: mem_type
    LOGICAL                                  :: dbg, docopy, &
                                                pad
    INTEGER(kind=int_8), DIMENSION(:), POINTER           :: newarray

!   ---------------------------------------------------------------------------
    !CALL timeset(routineN, error_handler)
    dbg = .FALSE.

    IF (PRESENT(array_resize)) NULLIFY(array_resize)

    IF (PRESENT (nocopy)) THEN
       docopy = .NOT. nocopy
    ELSE
       docopy = .TRUE.
    ENDIF
    IF (PRESENT (memory_type)) THEN
       mem_type = memory_type
    ELSE
       mem_type = dbcsr_memtype_default
    ENDIF
    lb_new = 1
    IF (PRESENT (lb)) lb_new = lb
    pad = .FALSE.
    IF (PRESENT (zero_pad)) pad = zero_pad
    !> Creates a new array if it doesn't yet exist.
    IF (.NOT.ASSOCIATED(array)) THEN
       IF(lb_new /= 1) &
          CPABORT("Arrays must start at 1")
       CALL mem_alloc_l (array, ub, mem_type=mem_type)
       IF (pad .AND. ub .GT. 0) CALL mem_zero_l (array, ub)
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    lb_orig = LBOUND(array,1)
    ub_orig = UBOUND(array,1)
    old_size = ub_orig - lb_orig + 1
    ! The existing array is big enough.
    IF (lb_orig.LE.lb_new .AND. ub_orig.GE.ub) THEN
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    ! A reallocation must be performed
    IF(dbg) WRITE(*,*)routineP//' Current bounds are',lb_orig,':',ub_orig,&
         '; special?' !,mem_type
    !CALL timeset(routineN,timing_handle)
    IF (lb_orig.GT.lb_new) THEN
       IF (PRESENT(factor)) THEN
          size_increase = lb_orig - lb_new
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          lb_new = MIN (lb_orig, lb_new - size_increase)
       ELSE
          lb_new = lb_orig
       ENDIF
    ENDIF
    IF (ub_orig.LT.ub) THEN
       IF (PRESENT(factor)) THEN
          size_increase = ub - ub_orig
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          ub_new = MAX (ub_orig, ub + size_increase)
       ELSE
          ub_new = ub
       ENDIF
    ELSE
       ub_new = ub
    ENDIF
    IF(dbg) WRITE(*,*)routineP//' Resizing to bounds',lb_new,':',ub_new
    !
    ! Deallocates the old array if it's not needed to copy the old data.
    IF(.NOT.docopy) THEN
       IF (PRESENT(array_resize)) THEN
          array_resize => array
          NULLIFY(array)
       ELSE
          CALL mem_dealloc_l (array, mem_type=mem_type)
       ENDIF
    ENDIF
    !
    ! Allocates the new array
    IF(lb_new /= 1) &
       CPABORT("Arrays must start at 1")
    CALL mem_alloc_l (newarray, ub_new-lb_new+1, mem_type)
    !
    ! Now copy and/or zero pad.
    IF(docopy) THEN
       IF(dbg .AND. (lb_new.GT.lb_orig .OR. ub_new.LT.ub_orig))&
            CPABORT("Old extent exceeds the new one.")
       IF (ub_orig-lb_orig+1 .gt. 0) THEN
          !newarray(lb_orig:ub_orig) = array(lb_orig:ub_orig)
          CALL mem_copy_l (newarray(lb_orig:ub_orig),&
               array(lb_orig:ub_orig), ub_orig-lb_orig+1)
       ENDIF
       IF (pad) THEN
          !newarray(lb_new:lb_orig-1) = 0
          CALL mem_zero_l (newarray(lb_new:lb_orig-1), (lb_orig-1)-lb_new+1)
          !newarray(ub_orig+1:ub_new) = 0
          CALL mem_zero_l (newarray(ub_orig+1:ub_new), ub_new-(ub_orig+1)+1)
       ENDIF
       IF (PRESENT(array_resize)) THEN
          array_resize => array
          NULLIFY(array)
       ELSE
          CALL mem_dealloc_l (array, mem_type=mem_type)
       ENDIF
    ELSEIF (pad) THEN
       !newarray(:) = 0
       CALL mem_zero_l (newarray, SIZE(newarray))
    ENDIF
    array => newarray
    IF (dbg) WRITE(*,*)routineP//' New array size', SIZE(array)
    !CALL timestop(error_handler)
  END SUBROUTINE ensure_array_size_l

! **************************************************************************************************
!> \brief Copies memory area
!> \param[out] dst   destination memory
!> \param[in] src    source memory
!> \param[in] n      length of copy
! **************************************************************************************************
  SUBROUTINE mem_copy_l (dst, src, n)
    INTEGER, INTENT(IN) :: n
    INTEGER(kind=int_8), DIMENSION(1:n), INTENT(OUT) :: dst
    INTEGER(kind=int_8), DIMENSION(1:n), INTENT(IN) :: src
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst,src)
    dst(:) = src(:)
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_copy_l

! **************************************************************************************************
!> \brief Zeros memory area
!> \param[out] dst   destination memory
!> \param[in] n      length of elements to zero
! **************************************************************************************************
  SUBROUTINE mem_zero_l (dst, n)
    INTEGER, INTENT(IN) :: n
    INTEGER(kind=int_8), DIMENSION(1:n), INTENT(OUT) :: dst
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst)
    dst(:) = 0
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_zero_l


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] n           length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_l (mem, n, mem_type)
    INTEGER(kind=int_8), DIMENSION(:), POINTER        :: mem
    INTEGER, INTENT(IN)                   :: n
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_l', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. n>1) THEN
       CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi .AND. dbcsr_data_allocation%use_mpi_allocator) THEN
!$omp critical(allocate)
       CALL mp_allocate(mem, n)
!$omp end critical(allocate)
    ELSE
       ALLOCATE(mem(n))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_l


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] sizes length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_l_2d (mem, sizes, mem_type)
    INTEGER(kind=int_8), DIMENSION(:,:), POINTER      :: mem
    INTEGER, DIMENSION(2), INTENT(IN)     :: sizes
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_l_2d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CPABORT("Accelerator hostalloc not supported for 2D arrays.")
       !CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CPABORT("MPI allocate not supported for 2D arrays.")
       !CALL mp_allocate(mem, n)
    ELSE
       ALLOCATE(mem(sizes(1), sizes(2)))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_l_2d


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_l (mem, mem_type)
    INTEGER(kind=int_8), DIMENSION(:), POINTER        :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_l', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. SIZE(mem)>1) THEN
       CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi .AND. dbcsr_data_allocation%use_mpi_allocator) THEN
       CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_l


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_l_2d (mem, mem_type)
    INTEGER(kind=int_8), DIMENSION(:,:), POINTER      :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_l', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CPABORT("Accelerator host deallocate not supported for 2D arrays.")
       !CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CPABORT("MPI deallocate not supported for 2D arrays.")
       !CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_l_2d


! **************************************************************************************************
!> \brief Sets a rank-2 pointer to rank-1 data using Fortran 2003 pointer
!>        rank remapping.
!> \param r2p ...
!> \param d1 ...
!> \param d2 ...
!> \param r1p ...
! **************************************************************************************************
  SUBROUTINE pointer_l_rank_remap2 (r2p, d1, d2, r1p)
    INTEGER, INTENT(IN)                      :: d1, d2
    INTEGER(kind=int_8), DIMENSION(:, :), &
      POINTER                                :: r2p
    INTEGER(kind=int_8), DIMENSION(:), &
      POINTER                                :: r1p

    r2p(1:d1,1:d2) => r1p(1:d1*d2)
  END SUBROUTINE pointer_l_rank_remap2
# 339 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr/data/dbcsr_ptr_util.f90"
# 152 "/builddir/build/BUILD/cp2k-6.1/src/dbcsr/data/dbcsr_ptr_util.F" 2

END MODULE dbcsr_ptr_util
